#!/usr/bin/env python
#
# Time-stamp: <00/09/09 16:52:10 dhellmann>
#
# COPYRIGHT
#
#   Permission to use, copy, modify, and distribute this software and
#   its documentation for any purpose and without fee is hereby
#   granted, provided that the above copyright notice appear in all
#   copies and that both that copyright notice and this permission
#   notice appear in supporting documentation, and that the name of
#   Doug Hellmann not be used in advertising or publicity pertaining
#   to distribution of the software without specific, written prior
#   permission.
#
# DISCLAIMER
#
#   DOUG HELLMANN DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
#   SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#   FITNESS, IN NO EVENT SHALL DOUG HELLMANN BE LIABLE FOR ANY
#   SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
#   AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
#   ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#   THIS SOFTWARE.
# 


"""Control class for writing out a set of documentation.

$Id: docset_multiplefile.py,v 1.5 2000/09/09 21:44:06 doughellmann Exp $
"""

__rcs_info__ = {
    #
    #  Creation Information
    #
    'module_name':'$RCSfile: docset_multiplefile.py,v $',
    'creator':'Doug Hellmann <doughellmann@home.com>',
    'project':'HappyDoc',
    'created':'Sun, 26-Mar-2000 11:19:54 EST',
    #
    #  Current Information
    #
    'author':'$Author: doughellmann $',
    'version':'$Revision: 1.5 $',
    'date':'$Date: 2000/09/09 21:44:06 $',
    'locker':'$Locker:  $',
    }

#
# Import system modules
#
import UserList
import sys
import os
import string
try:
    from cStringIO import StringIO
except:
    from StringIO import StringIO
import token
import symbol
import pprint

#
# Import Local modules
#


#
# Module
#

def entryPoint():
    "Return info about this module to the dynamic loader."
    return { 'name':'docset',
             'factory':DocSet,
             }

#
# Do not use multiple lines in the class docstring, since it shows up as
# part of the generated help message for the app and multiple lines throws
# off the formatting.
#
class DocSet(UserList.UserList):
    "DocSet where documentation is written to multiple files."

    def __init__(self,
                 formatterFactory,
                 list=None,
                 title='HappyDoc Generated Documentation',
                 baseDirectory=None,
                 statusMessageFunc=None,
                 description='',
                 appName='HappyDoc',
                 appVersion='',
                 appHome='',
                 author='',
                 formatterParameters={},
                 **extraNamedParameters
                 ):
        """Initialize the documentation set.

        Parameters:

            formatterFactory -- a callable object which creates the type
                                of formatter class to be used for formatting
                                the documentation set.  The object will be
                                called and passed the DocSet along with other
                                configuration values for the formatter.

            list -- an initial list of info objects from parseinfo.py

            title -- the title of the documentation set

            baseDirectory -- the name of the root directory where output
                             file(s) should be written

            statusMessageFunc -- function which will print a status message
                                 for the user

            description -- text describing the docset as a whole

            formatterParameters -- other named configuration values to be passed
                                   to the formatter when it is created through the
                                   factory.  Any unrecognized values will be
                                   quietly ignored.

        """        
        #
        # Store parameters
        #
        self._formatter_factory = formatterFactory
        self._title = title
        self._base_directory = baseDirectory
        self._status_message_func = statusMessageFunc
        self._docset_description = description
        self._formatter_configuration = formatterParameters
        #
        # Initialize this class
        #
        self._open_handles = []
        self._all_modules = {}
        self._all_classes = {}
        #
        # Initialize base class
        #
        UserList.UserList.__init__(self, list)
        self.statusMessage('Initializing documentation set %s...' % \
                           self._title)
        #
        # Handle unrecognized named parameters.
        #
        for extra_param, extra_value in extraNamedParameters.items():
            self.statusMessage(
                'WARNING: Parameter "%s" (%s) unrecognized by docset %s.' % \
                (extra_param, extra_value, self.__class__.__name__)
                )
        return

    def getBaseDirectory(self):
        "Returns the base directory for this documentation set."
        return self._base_directory

    def getClassInfo(self, className):
        "Returns class info if have it, None otherwise."
        return self._all_classes.get(className, None)

    def statusMessage(self, message=''):
        "Print a status message for the user."
        if self._status_message_func:
            self._status_message_func(message)
        return

    def write(self):
        "Write the documentation set to the output."
        self.statusMessage('Beginning to write...')
        #
        # Create the formatter
        #
        self._formatter = apply( self._formatter_factory,
                                 ( self, ),
                                 self._formatter_configuration )
        #
        # Get the name of and open the docset root file
        #
        self._root_name = self._formatter.getFullOutputNameForObject(None)
        self._root_node = self.openOutput( self._root_name, self._title, '' )
        self._formatter.writeText( self._docset_description, self._root_node )
        #
        # Write the output
        #
        self._writeModules()
        self._writeTOC()
        self._writeIndex()
        #
        # Close things up
        #
        self.close()
        return
    
    def _writeTOC(self):
        "Output the TOC."
        self.statusMessage()
        self.statusMessage('Writing table of contents...')
        #
        # Open a new section and list
        #
        self._formatter.pushSectionLevel(self._root_node)
        self._formatter.sectionHeader(self._root_node, 'Modules')
        #
        # Write module references to the list
        #
        #module_names = self._all_modules.keys()
        #module_names.sort()
        items = self._all_modules.items()
        module_ids = map(lambda x: (os.path.dirname(x[1].getFileName()),
                                    x[1].getFileName(),
                                    x[0]), items)
        module_ids.sort()
        modules_by_dir = {}
        for dir_name, file_name, module_name in module_ids:
            l = modules_by_dir.get(dir_name, [])
            l.append((file_name, module_name))
            modules_by_dir[dir_name] = l
        dirs = modules_by_dir.keys()
        dirs.sort()
        
        for dir_name in dirs:
            module_set = modules_by_dir[dir_name]
            if dir_name and (dir_name[-1] != '/'):
                dir_name = dir_name + '/'
            self._formatter.descriptiveListHeader(self._root_node,
                                                  dir_name)
            output_reduced_name = self._root_node.name[len(self._base_directory)+1:]
            for file_name, module_name in module_set:
                module = self._all_modules[module_name]
                self.statusMessage('\tAdding %s to TOC' % module_name)
                self._formatter.descriptiveListItem(
                    self._root_node,
                    self._formatter.getReference(module,
                                                 output_reduced_name),
                    module.getSummary()
                    )
            self._formatter.descriptiveListFooter(self._root_node)
        #
        # Close the section
        #
        self._formatter.popSectionLevel(self._root_node)
        return

    def _writeIndex(self):
        "Output the index."
        self.statusMessage()
        self.statusMessage('IMPLEMENT Writing index...')
        return

    def _writeModules(self):
        "Output documentation for all modules."
        self.statusMessage()
        self.statusMessage('Writing module documentation...')
        module_names = self._all_modules.keys()
        module_names.sort()
        for module_name in module_names:
            self._writeModule( module_name )
        return

    def _writeModule(self, module_name):
        "Output the documentation for the module named."
        module = self._all_modules[module_name]
        output_name = self._formatter.getFullOutputNameForObject(module)
        output = self.openOutput(output_name,
                                 'Module: %s' % module_name,
                                 module.getFileName())
        #
        # Write the doc string
        #
        self._formatter.writeText( module.getDocString(), output )
        #
        # Start the indented section
        #
        self._formatter.pushSectionLevel(output)
        #
        # List the dependant modules
        #
        imported_modules = module.getImportNames()
        if imported_modules:
            self._formatter.sectionHeader(output, 'Imported modules')
            self._formatter.listHeader(output, None, allowMultiColumn=0)
            imported_modules.sort()
            output_reduced_name = output.name[len(self._base_directory)+1:]
            for name in imported_modules:
                info = module.getImportInfo(name)
                if self._all_modules.has_key(name):
                    i_module = self._all_modules[name]
                    ref = self._formatter.getReference(
                        i_module,
                        output_reduced_name
                        )
                else:
                    ref = self._formatter.getPythonReference( name )
                    i_module = None
                if info:
                    if i_module:
                        #
                        # Process the list of imported names and
                        # generate references to the ones which are
                        # recognized by our parser.
                        #
                        import_list = []
                        for i_name in info:
                            try:
                                i_info = i_module.getClassInfo(i_name)
                                info_source = 'class'
                            except KeyError:
                                try:
                                    i_info = i_module.getFunctionInfo(i_name)
                                    info_source = 'function'
                                except KeyError:
                                    # do not have this name?
                                    import_list.append(i_name)
                                    continue

                            #
                            # Now that we have determined whether we have
                            # any information about the name, work out how
                            # to write the knowledge out.
                            #
                            output_reduced_name = output.name[len(
                                self._base_directory)+1:]
                            
                            if info_source == 'class':
                                #
                                # Get a reference to the file documenting the
                                # class.
                                #
                                i_ref = self._formatter.getReference(
                                    i_info,
                                    output_reduced_name,
                                    )
                                import_list.append(i_ref)
                                
                            elif info_source == 'function':
                                #
                                # Get a reference to the file documenting the
                                # module containing the function, and include
                                # the reference *into* the file to get right at
                                # the function.
                                #
                                i_ref = self._formatter.getNamedReference(
                                    i_module,
                                    i_name,
                                    output_reduced_name,
                                    )
                                import_list.append(i_ref)
                                
                            else:
                                #
                                # We do not know how to convert the name to a
                                # meaningful reference, so just put the name
                                # in the list to be output.
                                #
                                #print 'UNKNOWN TYPE: ', i_name
                                import_list.append(i_name)
                    else:
                        import_list = list(info)
                    self._formatter.listItem( output,
                                               'from %s import %s' % \
                                               ( ref, string.join(import_list, ', ') )
                                               )
                else:
                    self._formatter.listItem( output,
                                               'import %s' % ref
                                               )
            self._formatter.listFooter(output)
            self._formatter.sectionFooter(output)
        #
        # Write the info for the functions in this module
        #
        function_names = module.getFunctionNames()
        if function_names:
            self._formatter.sectionHeader( output, 'Functions' )
            function_names.sort()
            #
            # TOC list
            #
            self._formatter.listHeader( output )
            for function_name in function_names:
                self._formatter.listItem(
                    output,
                    self._formatter.getInternalReference(
                    module.getFunctionInfo(function_name)
                    )
                    )
            self._formatter.listFooter( output )
            #
            # Function descriptions
            #
            for function_name in function_names:
                self._writeFunction(function_name,
                                    module.getFunctionInfo,
                                    output)
            self._formatter.sectionFooter(output)
        #
        # Write the info for the classes in this module
        #
        class_names = module.getClassNames()
        if class_names:
            self._formatter.sectionHeader( output, 'Classes' )
            self._formatter.descriptiveListHeader( output, None )
            class_names.sort()
            output_reduced_name = output.name[len(self._base_directory)+1:]
            for class_name in class_names:
                c = module.getClassInfo(class_name)
                ref = self._formatter.getReference(c, output_reduced_name)
                self._formatter.descriptiveListItem(
                    output, ref, c.getSummary())
                class_output_name = self._formatter.getFullOutputNameForObject(c)
                class_output = self.openOutput(class_output_name,
                                               'Class: %s' % class_name,
                                               module.getFileName())
                self._writeClass( module, class_name, class_output )
            self._formatter.descriptiveListFooter(output)
            self._formatter.sectionFooter( output )
        #
        # Finish that indented level.
        #
        self._formatter.popSectionLevel(output)
        #
        # Close the output file
        #
        self.closeOutput(output)
        return


    
    def _writeBaseclassNames(self, parent, classInfo, output, indent=0):
        "Output the base class hierarchy for the given class."
        base_classes = classInfo.getBaseClassNames()
        if base_classes:
            if indent: self._formatter.indent(output)
            self._formatter.listHeader(output, None, allowMultiColumn=0)
            output_reduced_name = output.name[len(self._base_directory)+1:]
            for name in base_classes:
                try:
                    child = parent.getClassInfo(name)
                except KeyError:
                    self._formatter.listItem( output, name )
                else:
                    self._formatter.listItem(
                        output,
                        self._formatter.getReference(child,
                                                     output_reduced_name)
                        )
                    if name != classInfo.getName():
                        self._writeBaseclassNames(parent, child, output, 1)
            self._formatter.listFooter(output)
            if indent: self._formatter.dedent(output)
        return

    def _writeClass(self, parent, class_name, output):
        "Output the documentation for the class in the parent object."
        class_info = parent.getClassInfo(class_name)
        self._formatter.writeText( class_info.getDocString(), output)
        #
        # Base class hierarchy
        #
        base_classes = class_info.getBaseClassNames()
        if base_classes:
            self._formatter.pushSectionLevel(output)
            self._formatter.sectionHeader(output, 'Base Classes')
            self._writeBaseclassNames(parent, class_info, output)
            self._formatter.popSectionLevel(output)
        #
        # Start the indented section
        #
        self._formatter.pushSectionLevel(output)
        #
        # Write the info for the methods of this class
        #
        method_names = class_info.getMethodNames()
        if method_names:
            self._formatter.sectionHeader( output, 'Methods' )
            method_names.sort()
            #
            # TOC list
            #
            self._formatter.listHeader( output )
            for method_name in method_names:
                self._formatter.listItem(
                    output,
                    self._formatter.getInternalReference(
                    class_info.getMethodInfo(method_name)
                    )
                    )
            self._formatter.listFooter( output )
            for method_name in method_names:
                self._writeFunction(method_name,
                                    class_info.getMethodInfo,
                                    output)
            self._formatter.sectionFooter(output)
        #
        # Finish that indented level.
        #
        self._formatter.popSectionLevel(output)
        return

    def _writeFunction(self, function_name, getInfo, output):
        "Output the documentation for the function in the parent object."
        function = getInfo(function_name)
        #
        # Header
        #
        self._formatter.itemHeader( output, function )
        #
        # Function signature
        #
        self._formatter.writeFunctionSignature( output, function )
        #
        # Docstring
        #
        self._formatter.writeText( function.getDocString(), output )
        #
        # Exceptions
        #
        exception_names = function.getExceptionNames()
        if exception_names:
            self._formatter.pushSectionLevel(output)
            self._formatter.sectionHeader(output, 'Exceptions')
            self._formatter.writeExceptionListForFunction(output, function, None)
            self._formatter.popSectionLevel(output)
        return

    def close(self):
        "Close the open documentation set."
        for f in self._open_handles:
            try:
                self.closeOutput(f)
            except:
                pass
        return

    def append(self, infoObject):
        self._all_modules[ infoObject.getName() ] = infoObject
        for c in infoObject.getClassNames():
            self._all_classes[ c ] = infoObject.getClassInfo(c)
        UserList.UserList.append(self, infoObject)
        return

    def openOutput(self, name, title, subtitle):
        """
        Using this method to open output destinations
        means they will automatically be closed.
        """
        self.statusMessage(
            '\tDocumenting : "%s"' % title)
        f = self._formatter.openOutput(name, title, subtitle)
        self._open_handles.append(f)
        return f
    
    def closeOutput(self, output):
        "Close the output handle."
        self._formatter.closeOutput(output)
        return

    

    
