#!/usr/bin/env python
#
# Time-stamp: <00/09/09 18:53:50 dhellmann>
#
# COPYRIGHT
#
#   Permission to use, copy, modify, and distribute this software and
#   its documentation for any purpose and without fee is hereby
#   granted, provided that the above copyright notice appear in all
#   copies and that both that copyright notice and this permission
#   notice appear in supporting documentation, and that the name of Doug
#   Hellmann not be used in advertising or publicity pertaining to
#   distribution of the software without specific, written prior
#   permission.
# 
# DISCLAIMER
#
#   DOUG HELLMANN DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
#   INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
#   NO EVENT SHALL DOUG HELLMANN BE LIABLE FOR ANY SPECIAL, INDIRECT OR
#   CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#   OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
#   NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
#   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# 


"""Output documentation using HTML with tables.

$Id: formatter_htmlfile.py,v 1.6 2000/09/10 00:27:10 doughellmann Exp $


"""

__rcs_info__ = {
    #
    #  Creation Information
    #
    'module_name':'$RCSfile: formatter_htmlfile.py,v $',
    'creator':'Doug Hellmann <doughellmann@home.com>',
    'project':'Open Source',
    'created':'Sat, 03-Jun-2000 17:58:48 EDT',
    #
    #  Current Information
    #
    'author':'$Author: doughellmann $',
    'version':'$Revision: 1.6 $',
    'date':'$Date: 2000/09/10 00:27:10 $',
    }

#
# Import system modules
#
import os
import string

#
# Import Local modules
#
import formatter_file
import StructuredText
import indentstring

#
# Module
#

def entryPoint():
    "Return information about this module to the dynamic loader."
    return {
        'name':'htmltable',
        'factory':HTMLTableFormatter,
        }



class HTMLTableFormatter(formatter_file.FileBasedFormatter):
    'Output documentation using HTML with tables.'

    def __init__(self,
                 docSet, 
                 pageBackgroundColor='#ffffff',
                 levelOneHeadingBackgroundColor='#88bbee',
                 levelOneHeadingForegroundColor='#000000',
                 levelTwoHeadingBackgroundColor='#99ccff',
                 levelTwoHeadingForegroundColor='#000000',
                 appName='HappyDoc',
                 appVersion='Unknown',
                 appHome='http://happydoc.sourceforge.net',
                 docsetTitle=None,
                 dateStampFiles=1,
                 **configuration):
        '''Initialize the HTMLTableFormatter.

        Parameters

            'docSet' -- the DocSet instance containing global cross-reference
                      information

            'pageBackgroundColor' -- background color for HTML pages

            'levelOneHeadingBackgroundColor' -- background color for
            level one heading sections

            'levelOneHeadingForegroundColor' -- foreground color for
            level one heading sections

            'levelTwoHeadingBackgroundColor' -- background color for
            level two heading sections

            'levelTwoHeadingForegroundColor' -- foreground color for
            level two heading sections

            'dateStampFiles' -- Boolean indicating whether or not to
            include a date/time stamp in files.

            '**configuration' -- additional, optional, configuration values

        '''
        #
        # Preserve configuration parameters
        #
        self._page_bgcolor = pageBackgroundColor
        self._l1_bgcolor = levelOneHeadingBackgroundColor
        self._l1_fgcolor = levelOneHeadingForegroundColor
        self._l2_bgcolor = levelTwoHeadingBackgroundColor
        self._l2_fgcolor = levelTwoHeadingForegroundColor
        
        self._app_name = appName
        self._app_version = appVersion
        self._app_home = (appHome or 'http://www.python.org')

        self._date_stamp_files = int(dateStampFiles)
        
        #
        # Initialize the base class
        #
        apply(formatter_file.FileBasedFormatter.__init__, (self, docSet),
              configuration)
        return

    ##
    ## FileBasedFormatter implementation
    ##
    
    def getReference(self, infoSource, relativeSource):
        """Returns a reference to the 'infoSource' from 'relativeSource'.
        """
        #print '\ngetReference(', infoSource, ',', relativeSource, ')'
        link = computeRelativeLink(relativeSource,
                                   self.getOutputNameForObject(infoSource)
                                   )
        info = {
            'name':infoSource.getName(),
            'link':link,
            }
        ref = '<a href="%(link)s">%(name)s</a>' % info
        return ref
    
    def getNamedReference(self, infoSource, name, relativeSource):
        """Returns a reference to 'name' within the documentation for
        'infoSource' from 'relativeSource'.
        """
        #print '\ngetNamedReference(', infoSource.getName(), ',', name, ',', relativeSource, ')'
        link = computeRelativeLink(relativeSource,
                                   self.getOutputNameForObject(infoSource)
                                   )
        info = {
            'name':infoSource.getName(),
            'link':link,
            'target':name,
            }
        ref = '<a href="%(link)s#%(target)s">%(target)s</a>' % info
        return ref

    def getInternalReference(self, infoSource):
        """Returns a reference to 'infoSource' within the current document.
        """
        info = {
            'name':infoSource.getName(),
            }
        ref = '<a href="#%(name)s">%(name)s</a>' % info
        return ref
    
    def getPythonReference(self, moduleName):
        """Returns a reference to 'moduleName' documentation on the
        "Python.org":http://www.python.org documentation site.
        """
        if moduleName in self.sys_modules:
            return '<a href="http://www.python.org/doc/current/lib/module-%(moduleName)s.html">%(moduleName)s</a>' % locals()
        else:
            return moduleName
    
    def getFilenameExtension(self):
        "Returns the extension for creating output files."
        return 'html'

    def openOutput(self, name, title1, title2='&nbsp;'):
        """Open output destination using 'name' with the title from 'title1'.
        Write 'title2' as a secondary title to the new output.
        """
        f = formatter_file.FileBasedFormatter.openOutput(self,
                                                         name,
                                                         title1)
        self.htmlHeader( title1, title2,
                         self._l1_bgcolor,
                         self._l1_fgcolor,
                         f) 
        return f

    def closeOutput(self, output):
        "Close the 'output' handle."
        self.htmlFooter(output)
        output.close()
        return
    
    def pushSectionLevel(self, output):
        "Push a section level on the 'output' stack."
        self.writeRaw('''        
        <table border="0" cellpadding="5" cellspacing="0" width="100%%">
        ''', output)
        return

    def popSectionLevel(self, output):
        "Pop a section level from the 'output' stack."
        self.writeRaw('''
        </td>
        </tr>
        </table>
        ''', output)
        return


    def getRootLocation(self, output):
        "Return the root documentation node location relative to this 'output' location."
        first_file_opened = self.open_root_file.name
        current_output_name = output.name#[len(self._docset._base_directory)+1:]
        root_node_name = os.path.join(self._docset._base_directory,
                                      self.getRootNodeName())
        if first_file_opened == current_output_name:
            root_location = self.getRootNodeName()
            #print '**SAME'
        else:
            root_location = computeRelativeLink(current_output_name,
                                                root_node_name
                                                )
        return root_location


    def htmlHeader(self, title, subtitle, titleBg, titleFg, output):
        '''Output a standard HTML header used by all output files.

        Parameters

            'title' -- title of the document

            'output' -- destination for written output

            'titleBg' -- background color for the title bar

            'titleFg' -- foreground color for text in the title bar

        '''
        if not subtitle:
            subtitle = '&nbsp;'
        #
        # Determine where the root node is relative to the last
        # file opened.
        #
        root_location = self.getRootLocation(output)
            #print '**DIFFERENT'
        #
        # Put together parts of the header
        #
        info = {
            'bgcolor':self._page_bgcolor,
            'title':title,
            'subtitle':subtitle,
            'title_bg':titleBg,
            'title_fg':titleFg,
            'root':root_location,
            }
        self.writeRaw('''<html>

        <head>
        <title>%(title)s</title>
        </head>

        <body bgcolor="%(bgcolor)s">

        <p><i><a href="%(root)s">Table of Contents</a></i></p>
        
        <table border="0" cellpadding="5" cellspacing="0" width="100%%">
        <tr bgcolor="%(title_bg)s">
            <th rowspan="2"
                valign="top"
                align="left"
                width="10%%"><font color="%(title_fg)s">%(title)s</font>
            </th>
            <th align="right"><font color="%(title_fg)s">%(subtitle)s</font></th>
        </tr>
        <tr>
        <td>
        ''' % info, output)
        return
        

    def htmlFooter(self, output):
        "Output a standard HTML footer used by all 'output' files."
        if self._date_stamp_files:
            date_str = 'on %s' % self._update_time
        else:
            date_str = ''
        info = {
            'app_name':self._app_name,
            'app_version':self._app_version,
            'app_home':self._app_home,
            'date_str':date_str,
            'root':self.getRootLocation(output),
            }
        self.writeRaw('''
        </td>
        </tr>
        </table>

        <hr>

        <p><i><a href="%(root)s">Table of Contents</a></i></p>

        <i>This document was automatically generated %(date_str)s
        by <a href="%(app_home)s">%(app_name)s</a> version %(app_version)s</i>

        </body>
        </html>
        ''' % info, output)
        return

    def getRootNodeName(self):
        "Returns the name of the root node for documentation of this type."
        return 'index.html'

    ##
    ## HappyFormatterBase implementation
    ##

    def indent(self, output):
        "Begin an indented section."
        self.writeRaw('<ul>\n', output)
        return

    def dedent(self, output):
        "End an indented section."
        self.writeRaw('</ul>\n', output)
        return

    def writeText(self, text, output):
        "Format and write the 'text' to the 'output'."
        if not text:
            return
        text = self._unquoteString(text)
        #text = StructuredText.html_quote(text)
        html = StructuredText.HTML(text)
        self.writeRaw(str(html), output)
        return

    def formatCode(self, text):
        "Format 'text' as source code and return the new string."
        formatted_text = '<pre>\n%s\n</pre>' % StructuredText.html_quote(text)
        return formatted_text

    def writeCode(self, text, output):
        "Format and write the 'text' to 'output' as source code."
        if not text:
            return
        self.writeRaw(self.formatCode(text), output)
        return

    def listHeader(self, output, title=None, allowMultiColumn=1):
        """Output 'title' as a heading for a list.  If 'allowMultiColumn' is
        true, set up the list to have more than one column.
        """
        if title:
            self.writeRaw('<h4>%s</h4>\n' % title, output)
        self.writeRaw('\n', output)
        self._pushListContext(allowMultiColumn)
        return

    def listItem(self, output, text):
        "Format and output the 'text' as a list element."
        if self.current_list_context is not None:
            self.current_list_context.append(text)
        else:
            self.writeRaw('%s<br>\n' % text, output)
        return

    def _writeListItems(self, items, output):
        "Format and output the 'items' as list elements."
        #
        # Determine the geometry of the list
        # (number of columns and rows)
        #
        num_items = len(items)
        if num_items < 10:
            num_cols = 1
            num_rows = num_items
        elif num_items < 20:
            num_cols = 2
            num_rows = 10
        else:
            num_cols = 3
            if num_items < 30:
                num_rows = 10
            else:
                num_rows = (num_items / 3) + ((num_items % 3) / 3)
        #
        # Output the list
        #
        if num_cols == 1:
            for item in self.current_list_context:
                self.writeRaw('%s<br>\n' % item, output)
        else:
            self.writeRaw('<table border="0" cellspacing="2" cellpadding="2" width="100%">\n', output)
            self.writeRaw('<tr>', output)
            
            for col in range(num_cols):
                self.writeRaw('<td align="LEFT" valign="TOP">\n', output)
                base_item = col * num_rows
                for item_no in range(base_item, base_item + num_rows):
                    try:
                        self.writeRaw('%s<br>\n' % items[item_no], output)
                    except IndexError:
                        break
                self.writeRaw('</td>\n', output)

            self.writeRaw('</tr>', output)
            self.writeRaw('</table>\n', output)
        return
    
    def listFooter(self, output):
        "Write the closing footer for a list to the 'output'."
        if self.current_list_context:
            self._writeListItems(self.current_list_context, output)
        self.writeRaw('\n', output)
        self._popListContext()
        return

    def descriptiveListHeader(self, output, title):
        "Write the 'title' as the heading for a descriptive list to the 'output'."
        if title:
            self.writeRaw('<h4>%s</h4>\n' % title, output)
        self.writeRaw('<table border="0" cellpadding="5">\n', output)
        return

    def descriptiveListItem(self, output, item, description):
        "Format and write the 'item' and 'description' for a descriptive list to the 'output'."
        self.writeRaw('<tr><td valign="top" align="left">%s</td>' % item,
                      output)
        self.writeRaw('<td valign="top" align="left">', output)
        self.writeText(description, output)
        self.writeRaw('</td></tr>\n', output)
        return

    def descriptiveListFooter(self, output):
        "Write the closing footer for a descriptive list to the 'output'."
        self.writeRaw('</table>\n', output)
        return

    def genericSectionHeader(self, output, title1, title2, anchor=None):
        '''Output a standard nested table chunk which represents a section header.

        The output looks something like this::

            |--------|---------------------------|
            | title1 | title2                    |
            |        |---------------------------|
            |        | section text goes here
            |--------|

        Parameters

            'output' -- destination for written output

            'title1' -- title to be placed in left column

            'title2' -- title to be placed on top of right column

            'anchor' -- optional, anchor to which a reference can point
            to find this section

        '''
        if title1 is None:
            title1 = ''
        if title2 is None:
            title2 = ''
        bgcolor = '#cccccc'
        fgcolor = '#000000'
        info = {
            'bgcolor':self._l2_bgcolor,
            'fgcolor':self._l2_fgcolor,
            'title1':title1,
            'title2':title2,
            'anchor':anchor,
            }
        self.writeRaw('''
        <tr>
            <th bgcolor="%(bgcolor)s"
                rowspan="2"
                valign="top"
                align="left"
                width="20%%"
                >
                <font color="%(fgcolor)s">
                  <a name="%(anchor)s">%(title1)s</a>&nbsp;
                </font>
            </th>
            <th bgcolor="%(bgcolor)s"
                valign="top"
                align="left"
                >
                <font color="%(fgcolor)s">%(title2)s&nbsp;</font>
            </th>
        </tr>
        <tr>
        <td>
        ''' % info, output)
        return

    def genericSectionFooter(self, output):
        "Write a general purpose section closing footer to the 'output'."
        self.writeRaw('</td></tr>', output)
        return

    def sectionHeader(self, output, title):
        "Write a general purpose section openning title to the 'output'."
        self.genericSectionHeader( output, title, None, title )
        return
        
    def sectionFooter(self, output):
        "Write a general purpose section closing footer to the 'output'."
        self.genericSectionFooter( output )
        return

    def itemHeader(self, output, infoObject):
        "Write a section openning header for an 'infoObject' to the 'output'."
        name = infoObject.getName()
        self.genericSectionHeader( output, None, name, name )
        return

    def itemFooter(self, output):
        "Write a section closing footer to the 'output'."
        self.genericSectionFooter( output )
        return
        
    def dividingLine(self, output, fill='-'):
        "Write a sectional dividing line made up of repeated 'fill' characters to the 'output'."
        output.write('<hr>\n')
        return


def computeRelativeLink(fromName, toName):
    """Compute the relative link between fromName and toName.

    Both fromName and toName are strings refering to URL targets.
    This method computes the relative positions of the two nodes
    and returns a string which, if used as the HREF in a link in
    fromName will point directly to toName.
    """
    #print '\nFROM: ', fromName
    #print 'TO  : ', toName
    if fromName == toName:
        #print '\tsame name'
        relative_link = os.path.basename(toName)
    else:
        common_prefix = os.path.commonprefix( (os.path.dirname(fromName),
                                               os.path.dirname(toName))
                                              )
        from_name_no_prefix = fromName[len(common_prefix):]
        if from_name_no_prefix and from_name_no_prefix[0] == os.sep:
            from_name_no_prefix = from_name_no_prefix[1:]
        #print '\tfrom, no prefix:', from_name_no_prefix
        subdir_path = os.path.dirname(from_name_no_prefix)
        parts = string.split(subdir_path, os.sep)
        #print '\tparts:', parts
        if parts and parts[0]:
            levels = len(string.split(subdir_path, os.sep))
        else:
            levels = 0
        up_levels = (os.pardir + os.sep) * levels
        to_name_no_prefix = toName[len(common_prefix):]
        if to_name_no_prefix[0] == os.sep:
            to_name_no_prefix = to_name_no_prefix[1:]
        relative_link = "%s%s" % (up_levels, to_name_no_prefix)
    #print 'LINK: ', relative_link
    return relative_link
    
if __name__ == '__main__':
    for fro, to in (
        ('index.html', 'index.html'),
        ('HappyDoc/CommandLineApp.py.html', 'index.html'),
        ('HappyDoc/CommandLineApp.py_CommandLineApp.html', 'index.html'),
        ('HappyDoc/CommandLineApp.py_CommandLineApp.html',
         'HappyDoc/CommandLineApp.py.html'),
        ('/home/hellmann/devel/HappyDoc/doc/HappyDoc/ts_regex.py_compile.html',
         '/home/hellmann/devel/HappyDoc/doc/index.html'),
        
        ):
        #print 'FROM: ', fro
        #print 'TO  : ', to
        #print 'LINK: ', computeRelativeLink(fro, to)
        computeRelativeLink(fro, to)
        print
