#!/usr/bin/env python
#
# $Id: happydoc_class.py,v 1.12 2000/09/09 19:03:32 doughellmann Exp $
#
# Time-stamp: <00/09/09 14:36:51 dhellmann>
#
# Copyright Doug Hellmann 2000
#
#                         All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that both that copyright notice and this permission
# notice appear in supporting documentation, and that the name of Doug
# Hellmann not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission.
#
# DOUG HELLMANN DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
# NO EVENT SHALL DOUG HELLMANN BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

"""HappyDoc Class for HappyDoc Application.

"""

__rcs_info__ = {
    #
    #  Creation Information
    #
    'module_name'  : '$RCSfile: happydoc_class.py,v $',
    'rcs_id'       : '$Id: happydoc_class.py,v 1.12 2000/09/09 19:03:32 doughellmann Exp $',
    'creator'      : 'Doug Hellmann <doughellmann@bigfoot.com>',
    'project'      : 'HappyDoc',
    'created'      : 'Sun, 13-Aug-2000 11:27:00 EDT',

    #
    #  Current Information
    #
    'author'       : '$Author: doughellmann $',
    'version'      : '$Revision: 1.12 $',
    'date'         : '$Date: 2000/09/09 19:03:32 $',
}

#
# Import system modules
#
import pprint
import os
import glob
import sys
import types
import string
import re
import parser
import traceback

#
# Import Local modules
#
from parseinfo import getDocs
import parseinfo
from CommandLineApp import CommandLineApp
from prettyast import astListFixNames

#
# Module
#

import docset
import formatter
from cvsversion import cvs_product_version
    
class HappyDoc(CommandLineApp):
    """
    HappyDoc is a documentation generation/extraction tool which does
    not depend on being able to import modules.

    This app is based on the Demos/parser/example.py module distributed
    with the Python source distribution.

    """

    output_directory = './doc'
    output = None
    author_name = 'Doug Hellmann <doughellmann@bigfoot.com>'
    app_home = 'http://members.home.net/doughellmann/HappyDoc'
    package_description_file = 'README.txt'
    recurse_into_subdirs=1

    docset_type = None
    docset_title = 'HappyDoc Generated Documentation'

    #
    # Define the formats supported
    #
    supported_formats = formatter.plugins

    #
    # Define the documentation set types supported
    #
    supported_docset_types = docset.plugins
    
    def appInit(self):
        self._app_name = self.__class__.__name__
        self._app_version = cvs_product_version()
        self.statusMessage('%s version %s' % (self._app_name,
                                              self._app_version))
        self.set_docset_type('docset')
        self.set_format('htmltable')
        return

    def set_format(self, format):
        self.format = format
        try:
            self.formatter = self.supported_formats[format]
        except KeyError:
            raise ValueError('format must be one of %s' \
                             % self.supported_formats.keys(),
                             format)
        return

    def set_docset_type(self, docset_type):
        self.docset_type = docset_type
        try:
            self.docset = self.supported_docset_types[docset_type]
        except KeyError:
            raise ValueError('docset_type must be one of %s' % \
                             self.supported_docset_types.keys(),
                             docset_type)
        return
    
    def optionHandler_d(self, outputDirectory):
        "Specify an outputDirectory."
        self.output_directory = outputDirectory
        return

    def optionHandler_r(self):
        "Disable recursion into subdirectories."
        self.recurse_into_subdirs = 0
        return

    def optionHandler_o(self):
        "Specify that output should go to stdout."
        self.set_docset_type('stdout')
        return

    def optionHandler_t(self, title):
        "Specify a title for the documentation set."
        self.docset_title = title
        return

    def optionHandler_p(self, packageDescriptionFile):
        """Specify a file with a description of the package.

        The default packageDescriptionFile is README.txt.
        """
        self.package_description_file = packageDescriptionFile
        return

    def optionHandler_author(self, authorNameAndEmail):
        """Specify the author identification to be inserted for
        references.
        """
        self.author_name = authorNameAndEmail
        return

    def optionHandler_F(self, format):
        "Specify the output format."
        self.set_format(format)
        return

    def optionHandler_T(self, docset_type):
        "Specify the documentation set type."
        self.set_docset_type(docset_type)
        return

    def showVerboseSyntaxHelp(self):
        "Overloaded to show supported docset and format types."
        CommandLineApp.showVerboseSyntaxHelp(self)

        print 'SUPPORTED FORMATS for -F Option:\n'
        items = self.supported_formats.items()
        items.sort()
        for i in items:
            print '%20s: %s' % (i[0], i[1].__doc__)

        print '\nSUPPORTED DOCSET TYPES for -T Option:\n'
        items = self.supported_docset_types.items()
        items.sort()
        for i in items:
            print '%20s: %s' % (i[0], i[1].__doc__)

        print
        return

    
    def getFileInfo(self, fileName):
        "Parse the file and return the parseinfo instance for it."
        self.statusMessage('Getting info for %s' % fileName)
        return parseinfo.getDocs(fileName)
    
    def mkdir(self, fullname):
        """Make output directory as specified by 'path'.

        This function assumes that 'path' begins at the top of the
        filesystem and contains a completely specified pathname.  The
        path is made recursively, so all parent directories which do
        not exist should be created.
        
        """
        self.statusMessage('DEBUG: Checking for existance of %s' % fullname, 3)
        if not fullname:
            return
        if os.path.isdir(fullname) or os.path.islink(fullname):
            # fullname already exists, just return
            return
        (parent, filename) = os.path.split(fullname)
        #
        # Recursively create parent directory.
        #
        self.mkdir(parent)
        #
        # Make the subdirectory we decided didn't exist.
        #
        self.statusMessage('Creating directory %s' % fullname)
        os.mkdir(fullname)
        return
    
    def processFiles(self,
                     parsedModules,
                     fileNames,
                     useRecursion=1,
                     moduleFileName=re.compile(r'^.*\.py$').match,
                     IgnoreDirName=re.compile(r'^.*/(CVS|dist)$').match,
                     ):
        "Process a list of files."
        self.debugMethodEntry('processFiles',
                              parsedModules=parsedModules,
                              fileNames=fileNames,
                              useRecursion=useRecursion,)
        for file_name in fileNames:
            self.statusMessage('DEBUG: isdir(%s)=%s' % \
                               (file_name,
                                os.path.isdir(file_name)),
                               3)
            self.statusMessage('DEBUG: IgnoreDirName(%s)=%s' % \
                               (file_name,
                                IgnoreDirName(file_name)),
                               3)
            if ( os.path.isdir(file_name)
                 and
                 (not IgnoreDirName(file_name))
                 and
                 (useRecursion >= 0)
                 ):
                #
                # Find modules and directories within to
                # recurse.
                #
                dir_contents = glob.glob('%s/*' % file_name)
                self.statusMessage('Recursing into %s' % file_name)
                if not useRecursion:
                    useRecursion = -1
                self.processFiles(parsedModules=parsedModules,
                                  fileNames=dir_contents,
                                  useRecursion=useRecursion)
            elif moduleFileName(file_name):
                try:
                    self.statusMessage('DEBUG: Getting info on %s' % file_name, 3)
                    file_info = self.getFileInfo(file_name)
                except parser.ParserError, msg:
                    self.statusMessage('ERROR: %s' \
                                       % string.join(msg.args, ' '))
                    if self.verboseLevel >= 3:
                        traceback.print_exc()
                        print
                    
                else:
                    parsedModules.append(file_info)
            else:
                self.statusMessage('Skipping %s' % file_name, 3)
        return

    def getDocsetDescription(self, candidateFileName):
        """Get the description from the docset README file.

        If the candidateFileName file does not exist,
        returns an empty string.
        """
        docset_description = ''
        try:
            docset_description = open(candidateFileName, 'rt').read()
        except IOError:
            self.statusMessage('Could not get documentation set description from %s' \
                               % candidateFileName, 2)
        else:
            self.statusMessage('Got documentation set description from %s' \
                               % candidateFileName)
        return docset_description

    def getParameters(self, prefix, args, prefixSeparator='_'):
        """Find parameter settings in an argument sequence.
        """
        self.statusMessage('Looking for parameters with prefix "%s"' % prefix, 2)
        #
        # What we're going to return
        #
        parameter_set = {}
        ignored_values = []
        #
        # Local variables are faster to access
        #
        ignore = ignored_values.append
        find = string.find
        split = string.split
        prefix_len = len(prefix)
        full_prefix_len = prefix_len + len(prefixSeparator)
        #
        # Process candidates
        #
        for candidate in args:
            #
            # Check that this *is* a parameter
            #
            if find(candidate, '=') < 0:
                ignore(candidate)
                continue
            #
            # Check that this is our parameter
            #
            if candidate[:prefix_len] != prefix:
                ignore(candidate)
                continue
            #
            # Handle the parameter, it's ours.
            #
            param_with_prefix, val = split(candidate, '=')
            param = param_with_prefix[full_prefix_len:]
            self.statusMessage('DEBUG: Got %s parameter %s=%s' % (prefix, param, val), 3)
            parameter_set[param] = val
        return ignored_values, parameter_set
    
    def main(self, *args):
        #
        # Find DocSet arguments
        #
        args, docset_params = self.getParameters('docset', args)
        self.statusMessage('DEBUG: Docset parameters:', 3)
        for p, v in docset_params.items():
            self.statusMessage('DEBUG: \t%s:%s' % (p,v), 3)
        #
        # Find Formatter parameters
        #
        args, formatter_params = self.getParameters('formatter', args)
        formatter_params['appVersion'] = formatter_params.get('appVersion',
                                                              self._app_version)
        self.statusMessage('DEBUG: Formatter parameters:', 3)
        for p, v in formatter_params.items():
            self.statusMessage('DEBUG: \t%s:%s' % (p,v), 3)
        #
        # Get the list of modules to input
        #
        input_modules = args
        #
        # Create output directory
        #
        if not self.output:
            od = self.output_directory
            self.statusMessage('Output directory is %s' % self.output_directory, 2)
            if (od[0] == '.'):
                od = os.path.join( os.getcwd(), od )
                self.statusMessage('Setting output directory to %s' % od, 2)
            od = os.path.normpath(od)
            self.statusMessage('Creating output directory %s' % od, 2)
            self.mkdir(od)
            self.output_directory = od
        #
        # Get the general description text
        #
        self.statusMessage('Checking for package description file.', 2)
        if self.package_description_file:
            docset_description = None
            if (len(input_modules) == 1) and os.path.isdir(input_modules[0]):
                docset_description = self.getDocsetDescription(
                    os.path.join(input_modules[0], self.package_description_file)
                    )
            if not docset_description:
                docset_description = self.getDocsetDescription(
                    self.package_description_file)
        else:
            self.statusMessage('No documentation set description file specified, skipping')
            docset_description = ''
        #
        # Create the docset
        #
        docset_init_params = {
            'formatterFactory':self.formatter,
            'title':self.docset_title,
            'baseDirectory':self.output_directory,
            'statusMessageFunc':self.statusMessage,
            'appName':self._app_name,
            'appVersion':self._app_version,
            'appHome':self.app_home,
            'author':self.author_name,
            'description':docset_description,
            'formatterParameters':formatter_params,
            }
        docset_init_params.update(docset_params)
        parsed_modules = apply( self.docset, (), docset_init_params)
        #
        # Process the files and populate the docset
        #
        self.statusMessage('DEBUG: Input modules: %s' % input_modules, 3)
        self.processFiles(parsedModules=parsed_modules,
                          fileNames=input_modules,
                          useRecursion=self.recurse_into_subdirs)
        #
        # Tell the docset to output its results
        #
        parsed_modules.write()
        return

